import typer

from clearconf.utils.file import find_cconf_config, load_cconf_config, save_cconf_config
from clearconf.utils.stdout import print_dict

app = typer.Typer()

@app.command()
def add(script: str, config: str):
    if config[-3:] != '.py':
        print('Failed: Configurations have to be python files')
        return

    config = config[:-3]

    cconf_conf_path = find_cconf_config()
    cconf_conf = load_cconf_config(cconf_conf_path)

    cconf_conf['defaults'][script] = config
    save_cconf_config(cconf_conf, cconf_conf_path)


@app.command()
def remove(script):
    cconf_conf_path = find_cconf_config()
    cconf_conf = load_cconf_config(cconf_conf_path)

    cconf_conf['defaults'].pop(script)
    save_cconf_config(cconf_conf, cconf_conf_path)


@app.command()
def list():
    cconf_conf_path = find_cconf_config()
    cconf_conf = load_cconf_config(cconf_conf_path)

    print_dict(cconf_conf['defaults'])


import importlib
import __main__

from clearconf.utils.file import find_cconf_config, load_cconf_config
from clearconf.utils.conf import get_configs, get_default


cconf_conf_path = find_cconf_config()

if cconf_conf_path is None:
    raise ImportError("Couldn't find clearconf configuration file.\n"
          "Run 'cconf init' in your project root directory.")

cconf_conf = load_cconf_config(cconf_conf_path)
config_list = get_configs(cconf_conf)

try:
    usr_conf = get_default(cconf_conf, __main__.__file__.split('/')[-1])
    assert usr_conf is None or usr_conf[-3:] != '.py'
except AttributeError:
    usr_conf = None

if usr_conf is None:
    print('You can specify a default configuration with "cconf default"')
    for i, c in enumerate(config_list):
        print(f'{i}: {c}')

    print('Choose a configuration file:')
    k = input()
    usr_conf = config_list[int(k)]


Config = importlib.import_module(f'{__package__}.{usr_conf}').Config
Config.init()



from clearconf.utils.file import find_cconf_config, load_cconf_config
from pathlib import Path


def get_configs(cconf_conf):
    config_list = []

    # List all files ending with _cfg in the cfg_root directory
    project_root = Path(cconf_conf['project_root'])
    config_root = project_root / Path(cconf_conf['conf_dir'])

    for file in config_root.rglob('*'):

        if file.suffix == '.py' and file.stem[-5:] == '_conf':
            config_list.append(file.relative_to(config_root).as_posix()[:-3])

    return config_list


def get_default(cconf_conf, script_name):
    default_confs = cconf_conf['defaults']

    for script in default_confs:
        if script == script_name:
            return default_confs[script]

    return None



from .node import Node

class MetaBaseConfig(type):
    def __init__(cls, name, bases, clsdict):
        '''This executes after the configuration subclassing baseconfig
           is defined and it adds BaseConfig as a super class of all nested
           configurations'''
        super(MetaBaseConfig, cls).__init__(name, bases, clsdict)
        
        if len(cls.mro()) > 2:
            cls.__visit()
    
    def __visit(cls):
        from clearconf.api._utils.misc import (expand_name, add_parent, resolve_eval, subclass)
        from clearconf.api._utils.pickle_reduce import add_pickle_reduce

        cls._initialized = False
        
        cls._name = expand_name(cls)
        nodes = [Node(name, parent=cls) for name in dir(cls)]
        cls._nodes = nodes

        for node in nodes:
            subclass(node)
            add_parent(node) # if you add the parent before subclass an infinite loop happens somehow
            resolve_eval(node)
            # add_pickle_reduce(node)

        cls._initialized = True

    def __setattr__(cls, key, value):
        from clearconf.api._utils.misc import add_parent
        super().__setattr__(key, value)
        
        # This will let you add a baseconfig as an attribute of another while maintaining the tree
        if isinstance(value, MetaBaseConfig) and value._initialized and cls._initialized:
            node = Node(key, parent=cls, value=value)
            cls._nodes.append(node)
            add_parent(node)



"""This is an example of how a Config looks like.
    You can remove the parts that you don't need and add
    new sections as you please."""

from pathlib import Path
from clearconf import BaseConfig
from models import MLP
from datasets import ImageNet
from torch.optim import Adam
from torch.nn import BCEWithLogitsLoss

class Config(BaseConfig):
    name = Path(__file__).parts[-1]
    seed = 1234

    class Model(BaseConfig):
        architecture = MLP

        class Params(BaseConfig):
            num_layers = 16
            layers_dim = [96] * num_layers

    class Data(BaseConfig):
        input_dimension = 16384 * 2
        split = [0.1, 0.4, 0.5]
        noise_rate = 0.1
        tolerance = 0.001

        class DataSet(BaseConfig):
            dataset = ImageNet

            class Params(BaseConfig):
                root = './data/PCN'
                split = 'PCN.json'
                subset = 'train'
                length = None
                pick = [0]

        class DataLoader(BaseConfig):

            class Params(BaseConfig):
                batch_size = 1
                shuffle = False
                num_workers = 0
                pin_memory = True

    class Train(BaseConfig):
        loss_fn = BCEWithLogitsLoss()
        device = 'cuda'
        epochs = int(50_000)

        class Optim(BaseConfig):
            optim = Adam

            class Params(BaseConfig):
                lr = 0.0001

    class Logger(BaseConfig):
        active = False
        project = 'implicit-rep'
        task = 'input[0.1-0.4]-adam-(1)'


# if __name__ == '__main__':
#     Config.init()
#
#     from experiments import training as exp
#     exp.main()



class instantiate(type):
    def __new__(cls, clsname, bases, attrs):
        return type.__new__(cls, clsname, bases, attrs)('[eval]cfg')

class Root(str, metaclass=instantiate):
    
    def __new__(cls, string):
        instance = super().__new__(cls, string)
        return instance
    
    def __getattr__(self, name):
        return type(Root)(self + '.' + name)


import shutil
from pathlib import Path
import os

import clearconf
import typer

from clearconf.cli import defaults

from clearconf.utils.file import find_cconf_config, load_cconf_config, save_cconf_config
from clearconf.utils.stdout import print_list
from clearconf.utils.conf import get_configs

app = typer.Typer()
app.add_typer(defaults.app, name="defaults")

assets_path = Path(os.path.abspath(clearconf.__file__)).parent / "assets"


@app.command()
def init():
    cfg_root = Path('configs')
    if cfg_root.exists():
        print(f'Directory {cfg_root.as_posix()} already exists.')
    else:
        cfg_root.mkdir()
        shutil.copy(assets_path / 'init.py', cfg_root / '__init__.py')
        shutil.copy(assets_path / 'stub_conf.py', cfg_root)

    cconf_conf_path = Path('.clearconf')
    if cconf_conf_path.exists():
        print(f'File {cconf_conf_path.as_posix()} already exists.')
    else:
        cconf_conf_path.touch()

        cconf_conf = {
            'project_root': os.getcwd(),
            'conf_dir': cfg_root.as_posix(),
            'defaults': {}
        }

        save_cconf_config(cconf_conf, cconf_conf_path)

@app.command()
def list():
    cconf_conf_path = find_cconf_config()

    if cconf_conf_path is None:
        print("Couldn't find clearconf configuration file.\n"
              "Run 'cconf init' in your project root directory.")
        return

    cconf_conf = load_cconf_config(cconf_conf_path)
    config_list = get_configs(cconf_conf)

    print_list(config_list)


@app.callback()
def doc():
    """
    clearconf CLI can be used to initialized your
    project configurations.
    """


def main():
    app()





class _NestedClassGetter(object):
    # https://stackoverflow.com/questions/1947904/how-can-i-pickle-a-dynamically-created-nested-class-in-python/11493777#11493777
    # I can't reproduce the reason why this is required easily. However using torchdata with multiple workers can trigger it
    """
    When called with the containing class as the first argument, 
    and the name of the nested class as the second argument,
    returns an instance of the nested class.
    """
    def __call__(self, containing_class, class_name):
        nested_class = getattr(containing_class, class_name)

        # make an instance of a simple object (this one will do), for which we can change the
        # __class__ later on.
        nested_instance = _NestedClassGetter()

        # set the class of the instance, the __init__ will never be called on the class
        # but the original state will be set later on by pickle.
        nested_instance.__class__ = nested_class
        return nested_instance


def add_pickle_reduce(node):
    if not node.is_config or node.is_visited or node.is_hidden or node.is_private:
        return
    
    def _pickle_reduce(self):
        # return a class which can return this class when called with the 
        # appropriate tuple of arguments
        state = self.__dict__.copy()
        return (_NestedClassGetter(), (node.parent, self.__class__.__name__, ), state)
    setattr(node.value, '__reduce__', _pickle_reduce)


import __main__
import collections
import inspect
import json
import copy

from .meta import MetaBaseConfig

class BaseConfig(metaclass=MetaBaseConfig):

    @classmethod
    def to_json(cls):
        return json.dumps(cls.to_dict())

    # @classmethod
    # def to_dict2(cls):
    #     res = {}
        
    #     for node in cls._nodes():
            
    #         if node.is_config:
    #             res[node.name] = node.value.to_dict2()
                
            # property
            # function
                # lambda
            

    @classmethod
    def to_dict(cls, add_parent=False):
        target = cls

        # This add target as the top key of the dictionary
        #  if add_parent is True
        res = output = {}
        if add_parent:
            key = target.__name__
            if len(target.mro()) >= 4:
                key =  target.mro()[2].__name__
            if len(target.mro()) >= 5:
                key = f'{key}({target.mro()[3].__name__})'
            output = {key: res}

        target_attr = copy.deepcopy(dir(target))
        # This removes variables from superclasses
        for attr in [a for a in target.mro() 
                     if a.__name__ not in [target.__name__, f'{target.__name__}_mod', target.__name__[:-4], 'BaseConfig', 'Config', 'object']]:
            # The allows inheritance between configs but I guess there are better solutions
            if 'configs' not in attr.__module__:
                target_attr = list(filter(lambda x: x not in dir(attr), target_attr))

        for k in target_attr:
            if not k.startswith('_') and k not in ['to_dict', 'to_json', 'to_list', 'init', 'to_flat_dict', 'get_cfg', 'parent']:
                attr = getattr(target, k)
                
                if type(attr).__name__ == 'property':
                    attr = attr.fget

                # If it's a module get inside
                if hasattr(attr, '__module__'):

                    # If it's a class inside config, get inside it,
                    # else just log module and name in the dict as a string.

                    # if we are executing the config the module is __main__. If we are importing it is config
                    if type(attr).__name__ == 'function':
                        if attr.__name__ == '<lambda>':
                            funcString = str(inspect.getsourcelines(attr)[0])
                            res[k] = funcString.strip("['\\n']").split(" = ")[1]
                        else:
                            res[k] = f'function: {attr.__name__}'
                    elif attr.__module__.split('.')[0] == '__main__' or 'config' in attr.__module__:
                        if not inspect.isclass(attr): attr = type(attr)
                        
                        subclass_names = [a for a in [a.__name__ for a in attr.mro()] 
                                     if a not in [k, 'BaseConfig', 'object']]
                        
                        if len(subclass_names) > 0: # when a config class is subclassed to use it directly
                            k = f'{k}({subclass_names[0]})'

                        res[k] = attr.to_dict()
                    else:
                        # End up here if attr is not a class defined inside module.
                        if type(attr).__name__ == 'type':  # it is a class
                            name = f'{attr.__module__}.{attr.__name__}'
                        else: # it is an object
                            if attr.__str__ is not object.__str__:
                                name = attr.__str__()  # sometimes repr() might be preferable
                            else:
                                name = f'{type(attr).__name__}.{attr.__name__}'
                            res[k] = name
                # If it's not a class save it. This is done for basic types.
                # Could cause problems with complex objects
                else:
                    res[k] = attr

        return output

    @classmethod
    def to_flat_dict(cls) -> dict:
        import torch
        res = cls.to_dict()
        res = flatten(res)
        return res

    @classmethod
    def to_list(cls):
        target = cls

        res = []
        for k in dir(target):
            if not k.startswith('_') and k not in ['to_dict', 'to_json', 'to_list', 'init', 'to_flat_dict', 'get_cfg', 'parent']:
                attr = getattr(target, k)
                # If it's a class inside config, get inside it,
                # else just log module and name in the dict as a string
                if type(attr) == type:
                    if attr.__module__.split('.')[0] in ['configs', '__main__']:
                        res.append(attr.to_list())
                    else:
                        res.append(f'{attr.__module__}.{attr.__name__}')
                # If it's not a class save it. This is done for basic types.
                # Could cause problems with complex objects
                else:
                    res.append(attr)
        return res
    
    def get_cfg(self, to_self=False):
        if not to_self:
            return self.__class__
        
        target = self.__class__
        
        target_attr = dir(target)
        # This removes variables from superclasses
        for attr in [a for a in target.mro() 
                     if a.__name__ not in [target.__name__, target.__name__[:-4], 'BaseConfig', 'Config', 'object']]:
            # The allows inheritance between configs but I guess there are better solutions
            if 'configs' not in attr.__module__:
                target_attr = list(filter(lambda x: x not in dir(attr), target_attr))
        
        for k in target_attr:
            if not k.startswith('_') and k not in ['to_dict', 'to_json', 'to_list', 'init', 'to_flat_dict', 'get_cfg', 'parent']:
                attr = getattr(target, k)
                setattr(self, k, attr)
    
    # def __getattribute__(self, item):
    #     return object.__getattribute__(self, item)


def flatten(d, parent_key='', sep='.'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


"""This is an example of how a Config looks like.
    You can remove the parts that you don't need and add
    new sections as you please."""

from clearconf import BaseConfig


class Config(BaseConfig):
    pass



import pprint


def print_list(config_list: list):
    formatted_list = pprint.pformat(config_list, compact=True)

    remove_chr = ['[', ']', ',', '\'']

    for c in remove_chr:
        formatted_list = formatted_list.replace(c, '')

    print(formatted_list)


def print_dict(config_list: list):
    formatted_list = pprint.pformat(config_list, compact=True)

    remove_chr = ['{', '}', ',', '\'']

    for c in remove_chr:
        formatted_list = formatted_list.replace(c, '')

    print(formatted_list)


import inspect
import typing

from clearconf.api.hidden import Hidden


class Node:
    
    def __init__(self, name, parent=None, value=None) -> None:
        self._name = name
        self._alias = name
        
        if parent is None:
            parent = type('parent', (), {})
            setattr(parent, name, value)
        
        self.parent = parent
        
    @property
    def value(self):
        return getattr(self.parent, self._name)
        
    @value.setter
    def value(self, new):
        setattr(self.parent, self._name, new)
        
    @property
    def name(self):
        return self._alias
    
    @name.setter
    def name(self, alias):
        self._alias = alias
        
    @property
    def is_private(self):
        return self.name.startswith('_') or '._' in self.name

    @property
    def is_hidden(self):
        from clearconf.api.base_config import BaseConfig
        return (self.name in dir(BaseConfig) or typing.get_type_hints(self.parent).get(self.name, None) == Hidden or 
            Hidden in typing.get_args(typing.get_type_hints(self.parent).get(self.name, None)))
    @property      
    def is_visited(self):
        from clearconf.api.base_config import BaseConfig
        return issubclass(self.value, BaseConfig)
    
    @property
    def is_config(self):
        from clearconf.api.base_config import BaseConfig
        from clearconf.api._utils.misc import find_root
        # Attr is a class who has either been defined in the same module we are considering or is a
        # subclass of BaseConfig
        return (inspect.isclass(self.value) and 
            (find_root(self.value).__module__ == self.parent.__module__ or issubclass(self.value, BaseConfig)))
        
    def __repr__(self):
        return f'clearconf.Node({self.name})'

    # def __str__(self):
    #     sel


import os
from pathlib import Path
import json


def find_cconf_config():
    """
    Search for the .clearconf configuration file starting from the working directory and proceeding backwards.
    Stops looking the level before '/'.

    returns: pathlib.Path() of the .clearconf file
    """
    cconf_root = Path(os.getcwd())
    cconf_conf_path = None

    while cconf_root != Path('/'):
        cconf_conf_path = cconf_root / '.clearconf'

        if cconf_conf_path.exists():
            break

        cconf_conf_path = None
        cconf_root = cconf_root.parent

    return cconf_conf_path


def load_cconf_config(cconf_conf_path: Path):
    with cconf_conf_path.open('r') as f:
        cconf_conf = json.load(f)
    return cconf_conf


def save_cconf_config(cconf_conf: dict, cconf_conf_path: Path):
    with cconf_conf_path.open('w+') as f:
        json.dump(cconf_conf, f, indent=4)
        f.write('\n')




import copy
from functools import partial
from typing import Generic
from clearconf import BaseConfig, Hidden

def expand_name(target):
    superclasses = list(filter(lambda x: x not in [target.__name__, 'BaseConfig', 'object'], 
                        map(lambda x: x.__name__, target.mro())))
    if len(superclasses) > 0:
        return f'{target.__name__}:{superclasses[0]}'
    return target.__name__

# def expand_name(node):
#     if not node.is_config or node.is_hidden or node.is_private:
#         return
    
#     superclasses = list(filter(lambda x: x not in [node.name, 'BaseConfig', 'object'], 
#                         map(lambda x: x.__name__, node.value.mro())))

    # if len(superclasses) > 0:
    #     node.name =  f'{node.name}:{superclasses[0]}'
    # node.value.__name__ = node.name


def find_root(target):
    if not issubclass(target, BaseConfig):
        return target
    while True:
        try:
            target = target.parent
        except AttributeError:
            return target

def resolve(cls, body):
    cfg = find_root(cls)
    try:
        return eval(body)
    except Exception as e:
        # return 'Error in eval string evaluation'
        raise RuntimeError(f'The eval string {body} couldn\'t be resolved') from e

def resolve_eval(node):
    '''if the attribute is a string starting with [eval] the rest of the 
       string is evaluated and the result is substituted to the original
       attribute'''
    if node.is_private or node.is_hidden:
        return
       
    if isinstance(node.value, str) and node.value.startswith('[eval]'):
        body = copy.deepcopy(node.value[6:])
        node.value = classmethod(property(partial(resolve, body=body)))
    
def subclass(node):
    if not node.is_config or node.is_visited or node.is_private or node.is_hidden or node.name == 'parent':
        return
    
    if Generic in (base_classes := node.value.mro()): base_classes.remove(Generic) # necessary to avoid errors with typing
    # this create a new class equals to attr but which subclass BaseConfig
    node.value = type(node.name,
                 (BaseConfig, ) + tuple(base_classes),
                 dict(list(dict(vars(BaseConfig)).items()) + list(dict(vars(node.value)).items())))
    print()
    
def add_parent(node):
    if not node.is_config or node.is_hidden or node.is_private or node.name == 'parent':
        return
    
    node.value.parent: Hidden = node.parent